import os
import logging
import threading
import sys
import time

local = threading.local()
operateid_key = "operateid"


def set_operate_id(val):
    local.__setattr__(operateid_key, val)


class ThreadLocalFormatter(logging.Formatter):

    def format(self, record):
        msg = super(ThreadLocalFormatter, self).format(record)
        try:
            opid = local.__getattribute__(operateid_key)
            msg = opid + " " + msg
        except Exception:
            pass

        return msg


log_level = os.getenv("LOG_LEVEL", "info").upper()
# LOG_FORMAT = "%(asctime)s %(levelname)s - %(filename)s[:%(lineno)d] - %(message)s"
LOG_FORMAT = "%(asctime)s %(levelname)s %(message)s"
logging.Formatter.default_msec_format = '%s.%03d'
logging.Formatter.default_time_format = '%Y-%m-%dT%H:%M:%S'

stream_handler = logging.StreamHandler()
stream_handler.setFormatter(ThreadLocalFormatter(fmt=LOG_FORMAT))
# 直接设置datefmt毫秒时间戳无效 有疑问请查看logging.Formatter代码 不推荐使用handlers配置
# logging.basicConfig(level=log_level, format=LOG_FORMAT, handlers=[stream_handler])
logging.basicConfig(level=log_level, format=LOG_FORMAT)

# 重置handlers
for h in logging.root.handlers:
    logging.root.removeHandler(h)
logging.root.addHandler(stream_handler)

log = logging.getLogger("EP")


def run_time_decorator(title=""):
    def decorator(func):
        def wrapper(*args, **kwargs):
            # torch.cuda.synchronize()
            t0 = time.time()
            result = func(*args, **kwargs)
            # torch.cuda.synchronize()
            t1 = time.time()
            # print("{} call {} elapsed: {}ms ".format(title, func.__name__, (t1 - t0) * 1000))
            log.info("{}\t call {} \t elapsed:{:4.3f}ms \t".format(title, func.__name__, (t1 - t0) * 1000))
            return result

        return wrapper

    return decorator



# -*-coding: utf-8 -*-
"""
    @Author : panjq
    @E-mail : pan_jinquan@163.com
    @Date   : 2022-05-01 16:36:18
    @Url    : https://blog.csdn.net/xiaoyu_wu/article/details/102820384
"""

import threading
from queue import Queue
from typing import List
from concurrent.futures import ThreadPoolExecutor


class Base(object):
    def __init__(self, maxsize=5, *args, **kwargs):
        self.input: Queue = None  # 输入队列
        self.output: Queue = Queue(maxsize=maxsize)  # 输出队列

    def set_input(self, q: Queue):
        self.input = q

    def get_input(self):
        return self.input

    def set_output(self, q: Queue):
        self.output = q

    def get_output(self):
        return self.output

    def task(self, *args, **kwargs):
        """定义目标任务"""
        raise NotImplementedError("task not implemented!")

    def start(self, *args, **kwargs):
        """定义启动方法"""
        raise NotImplementedError("start not implemented!")

    def info(self):
        """打印信息"""
        info = ""
        if self.input:
            info += "input size:{}/{} ".format(self.input.qsize(), self.input.maxsize)
        if self.output:
            info += "output size:{}/{} ".format(self.output.qsize(), self.output.maxsize)
        info = "{} {}".format(self.__class__.__name__, info)
        return info


class Worker(Base):
    def __init__(self, maxsize=5, num_worker=2, *args, **kwargs):
        """
        :param input:  输入任务
        :param maxsize: 输出队列最大容量
        :param num_worker:  线程池最大线程数
        """
        super(Worker, self).__init__(maxsize=maxsize, args=args, kwargs=kwargs)
        self.num_worker = num_worker
        self.executor = ThreadPoolExecutor(num_worker)
        self.task_list = []

    def batch(self, batch_size, timeout=2):
        """
        :param batch_size: 获取一个batch的数据
        :param timeout: 超时
        :return: 
        """
        batch = []
        for i in range(batch_size):
            try:
                data = self.input.get(block=True, timeout=timeout)
                batch.append(data)
            except Exception as e:
                # print("{} 等待:{}".format(self.info(), e))
                break
        return batch

    def task(self, *args, **kwargs):
        """定义目标任务"""
        raise NotImplementedError("task not implemented!")

    def _target_v1(self):
        """
        输入和输出顺序一致
        :return:
        """
        while True:
            batch = self.batch(batch_size=self.num_worker)
            for r in self.executor.map(self.task, batch):
                self.output.put(r)

    def _target_v2(self, timeout=1):
        """
        输入和输出顺序不一致，先执行完的线程先输出,性能比_target_v1好
        :param timeout:
        :return:
        """
        while True:
            try:
                data = self.input.get(block=True, timeout=timeout)
                task = self.executor.submit(self.task, data)
                self.task_list.append(task)
            except Exception as e:
                pass
            finally:
                index = []
                for i, task in enumerate(self.task_list):
                    if task.done():
                        self.output.put(task.result(), block=True, timeout=None)
                    else:
                        index.append(i)
                self.task_list = [self.task_list[i] for i in index]

    def target(self):
        return self._target_v2()
        # return self._target_v1()

    def start(self, *args, **kwargs):
        t = threading.Thread(target=self.target, args=args, kwargs=kwargs)
        # 执行线程
        t.start()


class Compose(object):
    def __init__(self, workers: List[Worker]):
        self.workers = workers
        w0 = self.workers[0]
        for w in self.workers[1:]:
            w.set_input(w0.get_output())
            w0 = w

    def start(self, *args, **kwargs):
        for w in self.workers:
            w.start(*args, **kwargs)



# -*-coding: utf-8 -*-
"""
    @Author : panjq
    @E-mail : pan_jinquan@163.com
    @Date   : 2022-04-19 10:40:26
    @Brief  :
"""
import os
import toolz
import json
import numbers
from pybaseutils.file_utils import read_json_data, write_json_path
from typing import List, Tuple, Dict


class Dict2Obj(object):
    """ dict转类对象"""

    def __init__(self, args):
        self.__dict__.update(args)


def formatting(content):
    """格式化json数据"""
    info = json.dumps(content, indent=1, separators=(', ', ': '), ensure_ascii=False)
    return info


def get_keys_vaules(content, func=None):
    """
    遍历json数据并获得所有value的key路径
    :param content:
    :param func: 过滤条件函数func(k,v),默认为None,表示获取有的,获得所有value的key路径,一些常用的过滤方法：
           过滤所有文件：func = lambda k,v: isinstance(v, str) and os.path.isfile(v) and os.path.exists(v)
           过滤所有字符串：func = lambda k,v: isinstance(v, str)
           过滤所有数字：func = lambda k,v: isinstance(v, numbers.Number)
    :return: 返回满足条件的keys, values
    """

    def recursion(value, key=None, sub=[]):
        if not key is None: sub.append(key)
        if isinstance(value, list):
            for i in range(len(value)):
                recursion(value[i], key=i)
        elif isinstance(value, dict):
            for k, v in value.items():
                recursion(v, key=k)
        elif func is None:
            keys.append(sub.copy())
            values.append(value)
        elif func(key, value):
            keys.append(sub.copy())
            values.append(value)
        if sub: sub.pop()

    keys = []
    values = []
    recursion(content, key=None, sub=[])
    return keys, values


def get_value(content, key, default=None):
    """根据key路径获得对应的值"""
    value = toolz.get_in(key, content, default=default)
    return value


def get_values(content, keys):
    """根据keys路径获得对应的值"""
    values = [toolz.get_in(k, content) for k in keys]
    return values


def set_values(content, keys, values):
    """根据keys路径设置对应的值"""
    for k, v in zip(keys, values):
        content = toolz_assoc_in(content, keys=k, value=v)
        # data = toolz.assoc_in(data, keys=k, value=v)
    return content


def toolz_assoc_in(content, keys, value):
    """toolz_assoc_in用来代替toolz.assoc_in"""
    cur_keys = []
    for i, k in enumerate(keys):
        if isinstance(k, str):
            cur_keys.append(k)
        elif isinstance(k, int):
            curObj = toolz.get_in(cur_keys + [k], content)
            if curObj == None:
                print("发现非法参数:obj:{}, keys:{}".format(toolz.get_in(cur_keys, content), keys))
                raise Exception("给定路径非法")
            newKeys = keys[i + 1:]
            if len(newKeys) == 0:
                toolz.get_in(cur_keys, content)[k] = value
            else:
                newValue = toolz_assoc_in(curObj, newKeys, value)
                toolz.get_in(cur_keys, content)[k] = newValue
            return content
    if len(cur_keys) == len(keys):
        return toolz.assoc_in(content, cur_keys, value)


if __name__ == "__main__":
    content = {
        "code": "0",
        "data": {
            "image": ["image1", 0],
            "file": {"file1": "path/to/image1.jpg", "file2": 2, "file3": ["file3_v1", "file3_v2"], },
            "url": "url1"
        }
    }
    func = lambda k, v: isinstance(v, numbers.Number)
    # 遍历获得data中所有value的路径
    keys, values = get_keys_vaules(content, func=func)
    for k, v in zip(keys, values):
        print("path={}\t    value={}".format(k, v))
    print("===" * 20)
    # toolz使用toolz工具或得所有keys的值,values1与values的值是一样的
    # values1 = get_values(content, keys=keys)
    # values1 = get_values(content, keys=[['data1', 'image', 1], ['data', 'file', 'file11']])
    values1 = get_value(None, key=['data', 'image', 1],default={"data"})
    print(values1)
    print("===" * 20)
    values = list(range(len(values)))
    content = set_values(content, keys=keys, values=values)
    print(formatting(content))
    print("===" * 20)



# -*-coding: utf-8 -*-
"""
    @Author : panjq
    @E-mail : pan_jinquan@163.com
    @Date   : 2022-06-06 14:34:38
    @Brief  :
"""


def get_batch_sample(data_list, batch_size):
    """
    获得的一个batch-size的数据
    batch size data
    :param data_list:
    :param batch_size:
    :return:
    """
    sample_num = len(data_list)
    batch_num = (sample_num + batch_size - 1) // batch_size
    for i in range(batch_num):
        start = i * batch_size
        end = min((i + 1) * batch_size, sample_num)
        batch = data_list[start:end]
        yield batch


def get_batch_example(data_list, batch_size):
    """
    batch size data
    :param data_list:
    :param batch_size:
    :return:
    """
    sample_num = len(data_list)
    batch_num = (sample_num + batch_size - 1) // batch_size
    for i in range(batch_num):
        start = i * batch_size
        end = min((i + 1) * batch_size, sample_num)
        batch = data_list[start:end]
        print("batch:{}".format(batch))


if __name__ == "__main__":
    image_list = ["{}.jpg".format(i) for i in range(10)]
    batch_size = 15
    get_batch_example(image_list, batch_size)
    for batch in get_batch_sample(image_list, batch_size):
        print("batch:{}".format(batch))



# -*-coding: utf-8 -*-
"""
    @Author : panjq
    @E-mail : pan_jinquan@163.com
    @Date   : 2019-05-07 17:40:27
"""
import os

# windows: C:\Windows\Fonts可查看系统支持的字体
# Linux  : /usr/share/fonts/truetype可查看系统支持的字体

FONT_ROOT = os.path.dirname(__file__)
FONT_TYPE = {
    "楷体": "simkai.ttf",
    "宋体": "simsun.ttc",
    "仿宋": "simfang.ttf",
    "黑体": "simhei.ttf",
    # "simhei": "simhei.ttf", # 与黑体相近
    # "等线": "Deng.ttf",
    "方正": "方正粗黑宋简体.ttf",
}



# -*- coding: utf-8 -*-
"""
# --------------------------------------------------------
# @Author : panjq
# @E-mail : pan_jinquan@163.com
# @Date   : 2019-12-31 09:11:25
# --------------------------------------------------------
"""
import os
import cv2
import numpy as np
import PIL
from PIL import Image, ImageDraw, ImageFont
from pybaseutils import image_utils, file_utils
from pybaseutils.font_style import FONT_TYPE, FONT_ROOT


class FontType(object):
    def __init__(self, root=FONT_ROOT, style="宋体", size=20):
        """
        :param root: 安装字体的路径
        :param style: 选择字体风格
        :param size: 旋转字体大小
        """
        self.font_root = root
        self.font_style = style
        self.font_size = size
        self.font_file = os.path.join(self.font_root, FONT_TYPE[style])
        self.font_type = None
        if os.path.exists(self.font_file):
            self.font_type = ImageFont.truetype(self.font_file, size=self.font_size, encoding="utf-8")

    def set_root(self, root):
        """设置字体的路径"""
        self.__init__(root=root, style=self.font_style, size=self.font_size)

    def set_font_style(self, style, size):
        """设置字体风格和大小"""
        self.__init__(root=self.font_root, style=style, size=size)

    def set_font_type(self, font_file, size):
        self.font_root = os.path.dirname(font_file)
        self.font_style = os.path.basename(font_file)
        self.font_size = size
        self.font_type = ImageFont.truetype(self.font_file, size=size, encoding="utf-8")

    def get_font_type(self):
        return self.font_type


font_type = FontType()


def draw_image_text(image, point, text, style="楷体", size=20, color=(255, 255, 255)):
    """
    在图像中显示汉字
    https://blog.csdn.net/weixin_44237337/article/details/119817801
    :param image: RGB字体背景
    :param point:
    :param text:
    :param style: 字体风格
    :param size:
    :param color:
    :return:
    """
    image = Image.fromarray(image)
    draw = ImageDraw.Draw(image)
    if os.path.isfile(style):
        font = ImageFont.truetype(style, size=size, encoding="utf-8")
    else:
        # simhei.ttf 是字体，你如果没有字体，需要下载
        font_type.set_font_style(style=style, size=size)
        font = font_type.get_font_type()
    draw.text(point, text, fill=color, font=font)
    image = np.asarray(image)
    return image



def draw_font(text, style="楷体", scale=1.0, size=20, c1=(255, 255, 255), c2=(0, 0, 0), center=True):
    """
    绘制汉字
    :param text:
    :param style: 字体风格
    :param scale: 缩放因子
    :param size: 字体大小
    :param c1:字体颜色
    :param c2:背景颜色
    :param center: 是否居中显示
    :return:
    """
    if center:
        image = image_utils.create_image(shape=(size * 2, size * 2, 3), color=c2)
        image = draw_image_text(image, (0, 0), text, style=style, size=size, color=c1)
        if np.sum(image) < 1: return None
        mask = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
        box = image_utils.get_mask_boundrect_cv(mask, binarize=True, shift=10)
        image = image_utils.get_bbox_crop(image, box)
        image = image_utils.resize_image_padding(image, size=(size, size), color=(0, 0, 0))
    else:
        image = image_utils.create_image(shape=(size, size, 3), color=c2)
        image = draw_image_text(image, (0, 0), text, style=style, size=size, color=c1)
        if np.sum(image) < 1: return None
    if scale < 1.0:
        image = image_utils.get_scale_image(image, scale=scale, offset=(0, 0), color=c2)
    return image


if __name__ == "__main__":
    size = 512
    point = (0, 0)
    string = "我"
    for style, path in FONT_TYPE.items():
        style = "/home/dm/nasdata/dataset-dmai/ziku/1200款精品字体/书法字体库 37款/国祥手写体.ttf"
        # print(style, path)
        image = draw_font(string, style=style, size=size, scale=0.8, c2=(100, 0, 255))
        image_utils.cv_show_image("style", image, use_rgb=False)



# -*-coding: utf-8 -*-
"""
    @Author : panjq
    @E-mail : pan_jinquan@163.com
    @Date   : 2022-05-24 16:46:51
    @Brief  :
"""
import os
import cv2
import numpy as np
from tqdm import tqdm
from pybaseutils import image_utils, file_utils


def get_video_capture(video, width=None, height=None, fps=None):
    """
     获得视频读取对象
     --   7W   Pix--> width=320,height=240
     --   30W  Pix--> width=640,height=480
     720P,100W Pix--> width=1280,height=720
     960P,130W Pix--> width=1280,height=1024
    1080P,200W Pix--> width=1920,height=1080
    :param video: video file or Camera ID
    :param width:   图像分辨率width
    :param height:  图像分辨率height
    :param fps:  设置视频播放帧率
    :return:
    """
    video_cap = cv2.VideoCapture(video)
    # 设置分辨率
    if width:
        video_cap.set(cv2.CAP_PROP_FRAME_WIDTH, width)
    if height:
        video_cap.set(cv2.CAP_PROP_FRAME_HEIGHT, height)
    if fps:
        video_cap.set(cv2.CAP_PROP_FPS, 15)
    return video_cap


def get_video_info(video_cap: cv2.VideoCapture):
    """
    获得视频的基础信息
    :param video_cap:视频对象
    :return:
    """
    width = int(video_cap.get(cv2.CAP_PROP_FRAME_WIDTH))
    height = int(video_cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
    numFrames = int(video_cap.get(cv2.CAP_PROP_FRAME_COUNT))
    fps = int(video_cap.get(cv2.CAP_PROP_FPS))
    print("video:width:{},height:{},fps:{},numFrames:{}".format(width, height, fps, numFrames))
    return width, height, numFrames, fps


def get_video_writer(video_file, width, height, fps):
    """
    获得视频存储对象
    :param video_file: 输出保存视频的文件路径
    :param width:   图像分辨率width
    :param height:  图像分辨率height
    :param fps:  设置视频播放帧率
    :return:
    """
    if not os.path.exists(os.path.dirname(video_file)):
        os.makedirs(os.path.dirname(video_file))
    fourcc = cv2.VideoWriter_fourcc(*'XVID')
    frameSize = (int(width), int(height))
    video_writer = cv2.VideoWriter(video_file, fourcc, fps, frameSize)
    print("video:width:{},height:{},fps:{}".format(width, height, fps))
    return video_writer


def video2gif(video_file, gif_file=None, func=None, interval=1, use_pil=False, vis=True):
    """
    将视频文件直接转为GIF图像
    :param video_file: 输入视频文件
    :param gif_file: 保存GIF图文件
    :param func:
    :param interval:
    :param vis:
    :return:
    """
    name = os.path.basename(video_file).split(".")[0]
    if not gif_file:
        gif_file = os.path.join(os.path.dirname(video_file), name + ".gif")
    video_cap = get_video_capture(video_file)
    width, height, numFrames, fps = get_video_info(video_cap)
    if not os.path.exists(gif_file): file_utils.create_file_path(gif_file)
    count = 0
    frames = []
    while True:
        if count % interval == 0:
            # 设置抽帧的位置
            video_cap.set(cv2.CAP_PROP_POS_FRAMES, count)
            isSuccess, frame = video_cap.read()
            if not isSuccess:
                break
            if func:
                frame = func(frame)
            if vis:
                image_utils.cv_show_image("frame", frame, use_rgb=False, delay=30)
            frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
            frames.append(frame)
        count += 1
    video_cap.release()
    fps = fps / interval
    if use_pil:
        image_utils.frames2gif_by_pil(frames, gif_file=gif_file, fps=fps, loop=0)
    else:
        image_utils.frames2gif_by_imageio(frames, gif_file=gif_file, fps=fps, loop=0)


def video2frames(video_file, out_dir=None, func=None, interval=1, vis=True):
    """
    视频抽帧图像
    :param video_file: 视频文件
    :param out_dir: 保存抽帧图像的目录
    :param func: 回调函数，对每一帧图像进行处理
    :param interval: 保存间隔
    :param vis: 是否可视化显示
    :return:
    """
    name = os.path.basename(video_file).split(".")[0]
    if not out_dir:  out_dir = os.path.join(os.path.dirname(video_file), name)
    video_cap = get_video_capture(video_file)
    width, height, numFrames, fps = get_video_info(video_cap)
    if not os.path.exists(out_dir): os.makedirs(out_dir)
    count = 0
    while True:
        if count % interval == 0:
            # 设置抽帧的位置
            video_cap.set(cv2.CAP_PROP_POS_FRAMES, count)
            isSuccess, frame = video_cap.read()
            if not isSuccess:
                break
            if func:
                frame = func(frame)
            if vis:
                image_utils.cv_show_image("frame", frame, use_rgb=False, delay=30)
            frame_file = os.path.join(out_dir, "{}_{:0=4d}.jpg".format(name, count))
            cv2.imwrite(frame_file, frame)
        count += 1
    video_cap.release()


def frames2video(image_dir, video_file=None, func=None, postfix=["*.png", "*.jpg"], interval=1, fps=30, vis=True):
    """
    将抽帧图像转为视频文件(*.mp4)
    :param image_dir:抽帧图像路径
    :param video_file: 保存的视频文件
    :param func: 回调函数，对每一帧图像进行处理
    :param postfix:
    :param interval:
    :param fps:
    :param vis:
    :return:
    """
    image_list = file_utils.get_images_list(image_dir, postfix=postfix)
    if not video_file:
        video_file = os.path.join(image_dir) + "_{}.mp4".format(file_utils.get_time('p'))
    if len(image_list) == 0: return
    height, width = cv2.imread(image_list[0]).shape[:2]
    video_writer = get_video_writer(video_file, width, height, fps)
    for count, image_file in tqdm(enumerate(image_list)):
        if count % interval == 0:
            frame = cv2.imread(image_file)
            if func:
                frame = func(frame)
            if vis:
                image_utils.cv_show_image("frame", frame, use_rgb=False, delay=30)
            video_writer.write(frame)
    video_writer.release()


class CVVideo():
    def __init__(self):
        pass

    def start_capture(self, video_file, save_video=None, interval=1):
        """
        start capture video
        :param video_file: *.avi,*.mp4,...
        :param save_video: *.avi
        :param interval:
        :return:
        """
        # cv2.moveWindow("test", 1000, 100)
        video_cap = get_video_capture(video_file)
        width, height, numFrames, fps = get_video_info(video_cap)
        if save_video:
            self.video_writer = get_video_writer(save_video, width, height, fps)
        # freq = int(fps / detect_freq)
        count = 0
        while True:
            if count % interval == 0:
                # 设置抽帧的位置
                video_cap.set(cv2.CAP_PROP_POS_FRAMES, count)
                isSuccess, frame = video_cap.read()
                if not isSuccess:
                    break
                frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
                frame = self.task(frame)
            if save_video:
                self.write_video(frame)
            count += 1
        video_cap.release()

    def write_video(self, frame):
        self.video_writer.write(frame)

    def task(self, frame):
        # TODO
        cv2.imshow("image", frame)
        cv2.moveWindow("image", 0, 0)
        cv2.waitKey(10)
        return frame


def target_task(frame):
    frame = image_utils.resize_image(frame, size=(480, None))
    return frame


if __name__ == "__main__":
    # video_file = "../data/video/kunkun_cut.mp4"
    video_file = "/home/dm/nasdata/Demo/petal_20220524_164006.mp4"
    video_file = "/home/dm/nasdata/Demo/petal_20220524_183308.mp4"
    image_dir = "../data/video/kunkun_cut"
    # video2frames(video_file, interval=10, vis=True)
    # frames2video(image_dir, interval=1, vis=True)
    video2gif(video_file, interval=10, func=target_task, vis=True)



# -*-coding: utf-8 -*-
"""
    @Author : panjq
    @E-mail : pan_jinquan@163.com
    @Date   : 2022-06-29 18:31:12
    @Brief  :
"""
import os
import numpy as np
import cv2
import glob
import random
import numbers
import torch
import json
from tqdm import tqdm
from pybaseutils import image_utils, file_utils, coords_utils
from pybaseutils.dataloader.dataset import Dataset


class LabelMeDataset(Dataset):

    def __init__(self,
                 filename=None,
                 data_root=None,
                 anno_dir=None,
                 image_dir=None,
                 class_name=None,
                 use_rgb=True,
                 shuffle=False,
                 check=False,
                 **kwargs):
        """
        :param filename:
        :param data_root:
        :param anno_dir:
        :param image_dir:
        :param transform:
        :param use_rgb:
        :param shuffle:
        """
        super(LabelMeDataset, self).__init__()
        self.min_area = 1 / 1000  # 如果前景面积不足0.1%,则去除
        self.use_rgb = use_rgb
        self.class_name, self.class_dict = self.parser_classes(class_name)
        parser = self.parser_paths(filename, data_root, anno_dir, image_dir)
        self.data_root, self.anno_dir, self.image_dir, self.image_id = parser
        self.postfix = self.get_image_postfix(self.image_dir, self.image_id)
        self.classes = list(self.class_dict.values()) if self.class_dict else None
        self.num_classes = max(list(self.class_dict.values())) + 1 if self.class_dict else None
        self.class_weights = None
        if check:
            self.image_id = self.checking(self.image_id)
        if shuffle:
            random.seed(200)
            random.shuffle(self.image_id)
        self.num_images = len(self.image_id)
        self.scale_rate = 1.0
        self.target_type = 'gaussian'
        self.sigma = 2
        print("GridDataset class_name    :{}".format(class_name))
        print("GridDataset class_dict    :{}".format(self.class_dict))
        print("GridDataset num images    :{}".format(len(self.image_id)))
        print("GridDataset num_classes   :{}".format(self.num_classes))

    def __len__(self):
        return len(self.image_id)

    def parser_classes(self, class_name):
        """
        class_dict = {class_name: i for i, class_name in enumerate(class_name)}
        :param class_name:
                    str : class file
                    list: ["face","person"]
                    dict: 可以自定义label的id{'BACKGROUND': 0, 'person': 1, 'person_up': 1, 'person_down': 1}
        :return:
        """
        if isinstance(class_name, str):
            class_name = super().read_files(class_name)
        elif isinstance(class_name, numbers.Number):
            class_name = [str(i) for i in range(int(class_name))]
        elif isinstance(class_name, list) and "unique" in class_name:
            self.unique = True
        if isinstance(class_name, list):
            class_dict = {str(class_name): i for i, class_name in enumerate(class_name)}
        elif isinstance(class_name, dict):
            class_dict = class_name
            class_name = list(class_dict.keys())
        else:
            class_dict = None
        return class_name, class_dict

    def get_image_postfix(self, image_dir, image_id):
        """
        获得图像文件后缀名
        :param image_dir:
        :return:
        """
        if "." in image_id[0]:
            postfix = ""
        else:
            image_list = glob.glob(os.path.join(image_dir, "*"))
            postfix = os.path.basename(image_list[0]).split(".")[1]
        return postfix

    def get_image_anno_file(self, index):
        """
        :param index:
        :return:
        """
        image_id = self.index2id(index)
        image_file, annotation_file, image_id = self.__get_image_anno_file(self.image_dir,
                                                                           self.anno_dir,
                                                                           image_id,
                                                                           self.postfix)
        return image_file, annotation_file, image_id

    def __get_image_anno_file(self, image_dir, anno_dir, image_id: str, img_postfix):
        """
        :param image_dir:
        :param anno_dir:
        :param image_id:
        :param img_postfix:
        :return:
        """
        if not img_postfix and "." in image_id:
            img_postfix = image_id.split(".")[-1]
            image_id = image_id[:-len(img_postfix) - 1]
        image_file = os.path.join(image_dir, "{}.{}".format(image_id, img_postfix))
        annotation_file = os.path.join(anno_dir, "{}.json".format(image_id))
        return image_file, annotation_file, image_id

    def checking(self, image_ids: list, ignore_empty=True):
        """
        :param image_ids:
        :param ignore_empty : 是否去除一些空数据
        :return:
        """
        print("Please wait, it's in checking")
        dst_ids = []
        # image_ids = image_ids[:100]
        # image_ids = image_ids[100:]
        for image_id in tqdm(image_ids):
            image_file, annotation_file, image_id = self.get_image_anno_file(image_id)
            if not os.path.exists(annotation_file):
                continue
            if not os.path.exists(image_file):
                continue
            annotation = self.load_annotations(annotation_file)
            if len(annotation) == 0:
                continue
            dst_ids.append(image_id)
        print("have nums image:{},legal image:{}".format(len(image_ids), len(dst_ids)))
        return dst_ids

    def parser_paths(self, filename=None, data_root=None, anno_dir=None, image_dir=None):
        """
        :param filename:
        :param data_root:
        :param anno_dir:
        :param image_dir:
        :return:
        """
        if isinstance(data_root, str):
            anno_dir = os.path.join(data_root, "json") if not anno_dir else anno_dir
            image_dir = os.path.join(data_root, "images") if not image_dir else image_dir
        image_id = []
        if isinstance(filename, str):
            image_id = self.read_files(filename, split=",")
            data_root = os.path.dirname(filename)
        if not anno_dir:  # 如果anno_dir为空，则自动搜寻可能存在图片目录
            image_sub = ["json"]
            anno_dir = self.search_path(data_root, image_sub)
        if not image_dir:
            image_dir = self.search_path(data_root, ["JPEGImages", "images"])
        if anno_dir and not image_id:
            image_id = self.get_file_list(anno_dir, postfix=["*.json"], basename=True)
        elif image_dir and not image_id:
            image_id = self.get_file_list(anno_dir, postfix=["*.jpg"], basename=True)
        # assert os.path.exists(image_dir), Exception("no directory:{}".format(image_dir))
        # assert os.path.exists(anno_dir), Exception("no directory:{}".format(anno_dir))
        return data_root, anno_dir, image_dir, image_id

    def __getitem__(self, index):
        """
        :param index: int or str
        :return:rgb_image
        """
        image_id = self.index2id(index)
        image_file, json_file, image_id = self.get_image_anno_file(image_id)
        image = self.read_image(image_file, use_rgb=self.use_rgb)
        shape = image.shape
        annotation = self.load_annotations(json_file)
        box, label, point = self.parser_annotation(annotation, self.class_dict, shape)
        data = {"image": image, "point": point, "box": box, "label": label,
                "image_file": image_file, "json_file": json_file}
        return data

    def parser_annotation(self, annotation: dict, class_dict, shape=None):
        """
        :param annotation:  labelme标注的数据
        :param class_dict:  label映射
        :param shape: 图片shape(H,W,C),可进行坐标点的维度检查，避免越界
        :return:
        """
        bboxes, labels, points = [], [], []
        for anno in annotation:
            label = anno["label"].lower()
            if class_dict:
                if not label in class_dict:
                    continue
                if isinstance(class_dict, dict):
                    label = class_dict[label]
            pts = np.asarray(anno["points"], dtype=np.int32)
            if shape:
                h, w, d = shape
                pts[:, 0] = np.clip(pts[:, 0], 0, w - 1)
                pts[:, 1] = np.clip(pts[:, 1], 0, h - 1)
            box = image_utils.polygons2boxes([pts])[0]
            labels.append(label)
            bboxes.append(box)
            points.append(pts)
        return bboxes, labels, points

    def index2id(self, index):
        """
        :param index: int or str
        :return:
        """
        if isinstance(index, numbers.Number):
            image_id = self.image_id[index]
        else:
            image_id = index
        return image_id

    def __len__(self):
        return len(self.image_id)

    @staticmethod
    def get_files_id(file_list):
        """
        :param file_list:
        :return:
        """
        image_idx = []
        for path in file_list:
            basename = os.path.basename(path)
            id = basename.split(".")[0]
            image_idx.append(id)
        return image_idx

    def read_image(self, image_file: str, use_rgb=True):
        """
        :param image_file:
        :param use_rgb:
        :return:
        """
        image = cv2.imread(image_file, cv2.IMREAD_COLOR)
        if len(image.shape) == 2:
            image = cv2.cvtColor(image, cv2.COLOR_GRAY2BGR)
        elif image.shape[2] == 4:
            image = image[:, :, 0:3]
        if use_rgb:
            image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        return image

    def load_annotations(self, ann_file: str):
        with open(ann_file, "r") as f: annotation = json.load(f)
        annos = annotation["shapes"]
        return annos


def maker_labelme(json_file, points, labels, image_name, image_size, image_bs64=None):
    """
    制作label数据格式
    :param json_file: 保存json文件路径
    :param points: (num_labels,num_points,2)
    :param labels: (num_labels,)
    :param image_name: 图片名称，如果存在则进行拷贝到json_file同一级目录
    :param image_size: (W,H)
    :param image_bs64: 图片base64编码，可为None
    :return:
    """
    assert len(points) == len(labels)
    file_utils.create_file_path(json_file)
    shapes = []
    for point, label in zip(points, labels):
        # point = [[x1,y1],[x2,y2],...,[xn,yn]]
        if isinstance(point, np.ndarray): point = point.tolist()
        if not isinstance(point[0], list): point = [point]
        item = {"label": label, "line_color": None, "fill_color": None,
                "points": point, "shape_type": "polygon", "flags": {}}
        shapes.append(item)
    data = {
        "version": "3.16.7", "flags": {},
        "shapes": shapes,
        "lineColor": [0, 255, 0, 128],
        "fillColor": [255, 0, 0, 128],
        "imagePath": os.path.basename(image_name),
        "imageData": image_bs64,
        "imageHeight": image_size[1],
        "imageWidth": image_size[0]
    }
    if os.path.exists(image_name): file_utils.copy_file_to_dir(image_name, os.path.dirname(json_file))
    file_utils.write_json_path(json_file, data)
    return data


def show_target_image(image, bboxes, labels, points):
    image = image_utils.draw_image_bboxes_text(image, bboxes, labels, color=(255, 0, 0))
    image = image_utils.draw_landmark(image, points, color=(0, 255, 0))
    image_utils.cv_show_image("det", image)


if __name__ == "__main__":
    filename = "/home/dm/nasdata/dataset-dmai/handwriting/grid-det/grid_cross_points_v3/trainval.txt"
    # filename = None
    # anno_dir = "/media/dm/新加卷/SDK/base-utils/data/labelme"
    # image_dir = "/media/dm/新加卷/SDK/base-utils/data/labelme"
    input_size = [160, 160]
    dataset = LabelMeDataset(filename=filename,
                             data_root=None,
                             anno_dir=None,
                             image_dir=None,
                             class_name=None,
                             check=False,
                             phase="val",
                             shuffle=False)
    print("have num:{}".format(len(dataset)))
    for i in range(len(dataset)):
        print(i)  # i=20
        data = dataset.__getitem__(i)
        image, points, bboxes, labels = data["image"], data["point"], data["box"], data["label"]
        h, w = image.shape[:2]
        image_file = data["image_file"]
        json_file = os.path.join("masker", "{}.json".format(os.path.basename(image_file).split(".")[0]))
        maker_labelme(json_file, points, labels, (w, h), image_name=image_file, image_bs64=None)
        show_target_image(image, bboxes, labels, points)



# -*-coding: utf-8 -*-
"""
    @Author : panjq
    @E-mail : pan_jinquan@163.com
    @Date   : 2021-04-14 14:29:30
"""
import cv2
import numpy as np
import numpy.matlib


def create_gaussian_mask(width, height, center, radius=5):
    """
    生成二维高斯分布蒙版，width, height很大时，速度很慢
    :param width:
    :param height:
    :param center:
    :param radius: 高斯半径
    :return: gauss_mask : float32,(0,1)
    """
    # 利用for循环实现（很慢）
    # gauss_mask = np.zeros((height, width), dtype=np.float32)
    # for i in range(height):
    #     for j in range(width):
    #         dis = np.sqrt((i - center[1]) ** 2 + (j - center[0]) ** 2)
    #         gauss_mask[i, j] = np.exp(-0.5 * dis / radius)
    mask_x = np.matlib.repmat(center[0], height, width)
    mask_y = np.matlib.repmat(center[1], height, width)
    x1 = np.arange(width)
    x_map = np.matlib.repmat(x1, height, 1)
    y1 = np.arange(height)
    y_map = np.matlib.repmat(y1, width, 1)
    y_map = np.transpose(y_map)
    gauss_mask = np.sqrt((x_map - mask_x) ** 2 + (y_map - mask_y) ** 2)
    gauss_mask = np.exp(-0.5 * gauss_mask / radius)
    # gauss_mask = np.clip(gauss_mask, 0, 1)
    return gauss_mask


def get_image_resize(image, coordinates, dst_wh=[448, 448]):
    height, width, d = image.shape
    scale = [dst_wh[0] / width, dst_wh[1] / height]
    image = cv2.resize(image, dsize=tuple(dst_wh))
    coordinates = [[int(c[0] * scale[0]), int(c[1] * scale[1])] for c in coordinates]
    return image, coordinates


def get_coordinates_resize(coordinate, src_wh, dst_wh=[448, 448]):
    height, width = src_wh
    scale = [dst_wh[0] / width, dst_wh[1] / height]
    # coordinates = [[int(c[0] * scale[0]), int(c[1] * scale[1])] for c in coordinates]
    coordinate = [int(coordinate[0] * scale[0]), int(coordinate[1] * scale[1])]
    return coordinate


def fast_create_gaussian_mask(width, height, center, radius=2):
    """
    快速生成二维高斯分布蒙版
    :param width:
    :param height:
    :param center:
    :param radius: 高斯半径
    :return:
    """
    dst_wh = [128, 128]
    scale = [dst_wh[0] / width, dst_wh[1] / height]
    gauss_mask = create_gaussian_mask(int(width * scale[0]),
                                      int(height * scale[1]),
                                      [int(center[0] * scale[0]), int(center[1] * scale[1])],
                                      radius=radius)
    gauss_mask = cv2.resize(gauss_mask, dsize=(width, height))
    return gauss_mask


def get_image_heatmap(image, points: list, input_size=[448, 448], radius=2, fusion="gray"):
    """
    生成图像的热力图
    :param image: image array
    :param points: 坐标列表(N,2)
    :param radius: 半径
    :param fusion: 融合方式:[color,gray,split]
    :return:
    """
    # 生成热力图
    image, points = get_image_resize(image, points, dst_wh=input_size)
    h, w, d = image.shape
    mask = np.zeros((h, w), dtype=np.float32)
    for point in points:
        # mask += create_gaussian_mask(width=w, height=h, center=point, radius=radius)
        mask += fast_create_gaussian_mask(width=w, height=h, center=point, radius=radius)
    out_mask = np.clip(mask * 255.0, 0, 255)
    out_mask = np.asarray(out_mask, dtype=np.uint8)
    # 线性融合
    if fusion == "color":
        out_mask = cv2.applyColorMap(out_mask, cv2.COLORMAP_JET)
        heatmap = np.asarray(out_mask, dtype=np.float32)
        heatmap = heatmap * mask[:, :, np.newaxis]
        image = np.asarray(image, dtype=np.float32) + heatmap
        overlay = np.asarray(np.clip(image, 0, 255), dtype=np.uint8)
    elif fusion == "white":
        out_mask = cv2.cvtColor(out_mask, cv2.COLOR_GRAY2BGR)
        heatmap = np.asarray(out_mask, dtype=np.float32)
        image = np.asarray(image, dtype=np.float32) + heatmap
        overlay = np.asarray(np.clip(image, 0, 255), dtype=np.uint8)
    elif fusion == "black":
        out_mask = cv2.cvtColor(255 - out_mask, cv2.COLOR_GRAY2BGR)
        heatmap = np.asarray(out_mask, dtype=np.float32) / 255
        image = np.asarray(image, dtype=np.float32) * heatmap
        overlay = np.asarray(np.clip(image, 0, 255), dtype=np.uint8)
    elif fusion == "split":
        overlay = image.copy()
        overlay[:, :, 1] = mask
        out_mask = mask
    else:
        raise Exception("fusion ERROR:{}".format(fusion))
    # dst = src1[i] * alpha + src2[i] * beta + gamma;//两张图片每个通道对应数值之和。
    # 根据mask进行融合
    # ret, bmask = cv2.threshold(mask, 15, 255, cv2.THRESH_BINARY)
    # img1_bg = cv2.bitwise_and(mask, mask, mask=bmask)
    # img1_bg = cv2.cvtColor(img1_bg, cv2.COLOR_GRAY2BGR)
    # overlay = cv2.add(img1_bg, image)  # 进行融合
    return overlay, out_mask


if __name__ == "__main__":
    from pybaseutils import image_utils

    image_path = "/home/dm/project/python-learning-notes/dataset/test_image/test1.jpg"
    image = image_utils.read_image(image_path)
    input_size = [416, 416]
    points = [[100, 200], [400, 200]]
    colormap = "color"
    # image = np.zeros_like(image, dtype=np.uint8)
    image, mask = get_image_heatmap(image, points, input_size=input_size, fusion=colormap)
    image_utils.cv_show_image("mask", mask, delay=1)
    image_utils.cv_show_image("image", image)



# -*-coding: utf-8 -*-
"""
    @Author : panjq
    @E-mail : pan_jinquan@163.com
    @Date   : 2022-05-25 18:19:34
    @Brief  :
"""
import time
import numpy as np
from typing import Dict, List


class Recorder(object):
    def __init__(self):
        self.content: Dict = {}

    def push(self, key, v):
        if key in self.content:
            self.content[key]["total"] = self.content[key]["total"] + v
            self.content[key]["count"] = self.content[key]["count"] + 1
            self.content[key]["avg"] = self.content[key]["total"] / self.content[key]["count"]
        else:
            self.content[key] = {"avg": v, "total": v, "count": 1}

    def get(self, key) -> Dict:
        return self.content[key]

    def reset(self):
        self.content: Dict = {}

    def info(self, key):
        print(self.get(key))


recorder = Recorder()


def performance(tag=""):
    def decorator(func):
        def wrapper(*args, **kwargs):
            # torch.cuda.synchronize()
            t0 = time.time()
            result = func(*args, **kwargs)
            # torch.cuda.synchronize()
            t1 = time.time()
            key = tag if tag else str(func.__name__)
            recorder.push(key=key, v=(t1 - t0) * 1000)
            content = recorder.get(key)
            elapsed = "avg:{:.5f}ms\t total:{:.5f}ms\t count:{}".format(content["avg"],
                                                                        content["total"],
                                                                        content["count"])
            print("{} call {} elapsed: {}".format(tag, func.__name__, elapsed))
            return result

        return wrapper

    return decorator


class Performance(object):
    def __init__(self, tag=""):
        self.tag = tag

    def __enter__(self):
        self.t0 = time.time()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):  # 参数为异常信息
        self.t1 = time.time()
        recorder.push(key=self.tag, v=(self.t1 - self.t0) * 1000)
        self.info(key=self.tag)

    def info(self, key):
        content = recorder.get(key)
        elapsed = "avg:{:.5f}ms\t total:{:.5f}ms\t count:{}".format(content["avg"],
                                                                    content["total"],
                                                                    content["count"])
        print("{} elapsed: {}".format(self.tag, elapsed))

    def task(self):
        pass


@performance("")
def targe_func1():
    time.sleep(1)


@performance("")
def targe_func2():
    time.sleep(0.5)


def targe_func3():
    with Performance("targe_func3") as p:
        time.sleep(1)


def targe_func4():
    with Performance("targe_func4") as p:
        time.sleep(0.5)


def targe_func():
    targe_func1()
    targe_func2()
    targe_func3()
    targe_func4()


if __name__ == '__main__':
    for i in range(10):
        targe_func()



# -*-coding: utf-8 -*-
"""
    @Author : panjq
    @E-mail : pan_jinquan@163.com
    @Date   : 2022-03-22 09:11:35
    @Brief  :
"""
import sys
import os
import cv2
import base64
import numpy as np
from typing import Any

IMG_PREFIX = "image/png"  # 图片base64字符串前缀
precision = 6  # 小数点精度


def isbase64(data: str, prefix=IMG_PREFIX):
    """判断是否是二进制字符串图像"""
    return prefix == data[0:len(prefix)]


def base642image(bs64, prefix=IMG_PREFIX, use_rgb=False) -> np.ndarray:
    """
    將二进制字符串解码为图像
    :param bs64: 二进制字符串图像
    :param prefix: base64字符串前缀,用于表识字符串的类型
    :param use_rgb: True:返回RGB的图像, False:返回BGR格式的图像
    :return: 返回图像
    """
    if prefix == bs64[0:len(prefix)]:
        bs64 = bs64[len(prefix):]
    bs64 = bytes(bs64, 'utf-8')
    image = base64.b64decode(bs64)
    image = np.fromstring(image, np.uint8)
    image = cv2.imdecode(image, flags=cv2.IMREAD_UNCHANGED)
    # image = cv2.imdecode(image, cv2.IMREAD_COLOR)
    if use_rgb:
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
    return image


def image2base64(image: np.ndarray, prefix=IMG_PREFIX, use_rgb=False) -> str:
    """
    将图像编码为二进制字符串
    ``` python
        from io import BytesIO
        bgr_img = Image.fromarray(image)
        buff = BytesIO()
        mg.save(buff, format="PNG")
        image_base64 = base64.b64encode(buff.getvalue()).decode("utf-8")
        image_base64 = str(base64.b64encode(image), encoding='utf-8')
    ```
    :param image: 图像
    :param prefix: base64字符串前缀,用于表识字符串的类型
    :param use_rgb: True:输入image是RGB的图像, False:输入image是BGR格式的图像
    :return: 返回图像
    """
    img = image.copy()
    if len(img.shape) == 3 and use_rgb:
        img = cv2.cvtColor(img, cv2.COLOR_RGB2BGR)
    ext = prefix.split("/")
    ext = "." + ext[1] if len(ext) == 2 else ".png"
    img = cv2.imencode(ext, img)[1]
    bs64 = prefix + base64.b64encode(img).decode()
    return bs64


def file2base64(file, prefix=IMG_PREFIX) -> str:
    """
    将文件编码为base64字符串
    :param file: 文件路径
    :param prefix: base64字符串前缀,用于表识字符串的类型
    :return:base64字符串
    """
    bs64 = prefix + base64.b64encode(open(file, 'rb').read()).decode()
    return bs64


def base642file(file, bs64, prefix=IMG_PREFIX) -> str:
    """
    将base64字符串解码为文件
    :param file: 文件路径
    :param prefix: base64字符串前缀,用于表识字符串的类型
    :return:文件路径
    """
    if prefix == bs64[0:len(prefix)]:
        bs64 = bs64[len(prefix):]
    bs64 = base64.b64decode(bs64)
    with open(file, 'wb') as f: f.write(bs64)
    return file


def array2base64(data: Any, prefix=IMG_PREFIX, use_rgb=False) -> Any:
    """
    序列化:将输入数据含有图像数据(ndarray)都编码为base64字符串
    :param data: 输入数据
    :param prefix: base64字符串前缀,用于表识字符串的类型
    :param use_rgb: True:输入image是RGB的图像, False:输入image是BGR格式的图像
    :return:
    """
    if isinstance(data, np.ndarray) and data.dtype == np.uint8:
        return image2base64(data, prefix=prefix, use_rgb=use_rgb)
    elif isinstance(data, np.ndarray):
        return data.tolist()
    elif isinstance(data, np.integer):
        return int(data)
    elif isinstance(data, np.floating):
        return round(float(data), precision)
    elif isinstance(data, float):
        return round(data, precision)
    elif isinstance(data, list):
        for i in range(len(data)):
            data[i] = array2base64(data[i], prefix=prefix, use_rgb=use_rgb)
    elif isinstance(data, dict):
        for k, v in data.items():
            data[k] = array2base64(v, prefix=prefix, use_rgb=use_rgb)
    return data


def base642array(data: Any, prefix=IMG_PREFIX, use_rgb=False) -> Any:
    """
    反序列化:将输入数据含有base64字符串都解码为图像数据(ndarray)
    :param data: 输入数据
    :param prefix: base64字符串前缀,用于表识字符串的类型
    :param use_rgb: True:输入image是RGB的图像, False:输入image是BGR格式的图像
    :return:
    """
    if isinstance(data, str) and prefix == data[0:len(prefix)]:
        return base642image(data, prefix=prefix, use_rgb=use_rgb)
    elif isinstance(data, list):
        for i in range(len(data)):
            data[i] = base642array(data[i], prefix=prefix, use_rgb=use_rgb)
    elif isinstance(data, dict):
        for k, v in data.items():
            data[k] = base642array(v, prefix=prefix, use_rgb=use_rgb)
    return data


serialization = array2base64  # 序列化
deserialization = base642array  # 反序列化

if __name__ == "__main__":
    file = "/home/dm/project/python-learning-notes/utils/test.jpg"
    bgr1 = cv2.imread(file)
    image_base64 = image2base64(bgr1)
    image_base64 = file2base64(file)
    bgr2 = base642image(image_base64, use_rgb=False)



# -*-coding: utf-8 -*-
"""
    @Author : panjq
    @E-mail : pan_jinquan@163.com
    @Date   : 2022-03-29 16:15:09
    @Brief  :
"""
import cv2
import numpy as np
from typing import List, Tuple, Dict
from pybaseutils import image_utils, json_utils, file_utils


def concat_stroke_image(mask, seg_list, split_line=False, vis=False):
    """
    水平拼接笔画图片
    :param mask: 整字的笔画mask
    :param seg_list: 字的笔画mask列表
    :param split_line: 是否显示分隔线
    :param vis: 是否可视化
    :return:返回水平拼接笔画图片
    """
    if len(seg_list) > 0:
        seg_mask = np.max(seg_list, axis=0)
    else:
        seg_mask = np.zeros_like(mask)
    diff = np.abs(mask - seg_mask)
    images = [mask] + [seg_mask] + [diff] + seg_list
    vis_image = image_utils.image_hstack(images, split_line=split_line)
    if vis:
        image_utils.cv_show_image("mask-seg-diff-stroke", vis_image, use_rgb=False)
    return vis_image


def concat_hw_gt_stroke_image(hw_mask, hw_segs, gt_mask, gt_segs, split_line=True, vis=False):
    """
    对比标准字的笔画和手写字的笔画分割图
    :param hw_mask:手写字整字mask
    :param hw_segs:手写字分割后的笔画mask列表
    :param gt_mask:标准字整字mask
    :param gt_segs:标准字真实的笔画mask列表
    :param split_line: 是否显示分隔线
    :param vis: 是否可视化
    :return:
    """
    hw_stroke = concat_stroke_image(hw_mask, hw_segs, split_line=split_line, vis=False)
    gt_stroke = concat_stroke_image(gt_mask, gt_segs, split_line=split_line, vis=False)
    vis_image = image_utils.image_vstack([gt_stroke, hw_stroke], split_line=split_line)
    if vis:
        image_utils.cv_show_image("gt-pd-stroke", vis_image, use_rgb=False)
    return vis_image


def show_word_info(word_info):
    print(json_utils.formatting(word_info['content'] if 'content' in word_info else ""))
    keys = ["label", "stroke_label", "stroke_names"]
    content = ["{}\t:{}".format(key, word_info[key]) for key in keys if key in word_info]
    print("\n".join(content))
    if len(word_info['stroke_segs']) == 0 or word_info['mask'] is None: return
    stroke_img = concat_stroke_image(word_info['mask'], word_info['stroke_segs'], split_line=True, vis=False)
    image_utils.cv_show_image("mask-seg-stroke", stroke_img, use_rgb=False)
    for i in range(len(word_info['piece_segs'])):
        stroke = word_info['stroke_segs'][i]
        piece = word_info['piece_segs'][i]
        names = word_info['stroke_label'][i]
        print("i={:0=3d}\tname={}\tpiece_num={}".format(i, names, len(piece)))
        if len(piece) == 0: continue
        piece_img = concat_stroke_image(stroke, piece, split_line=True, vis=False)
        image_utils.cv_show_image("mask-seg-piece", piece_img, use_rgb=False)
    cv2.destroyWindow("mask-seg-piece")
    print("---" * 20)


def show_hw_gt_word_info(word_info):
    """可视化标准字和手写字的分割效果"""
    keys = ["label", "stroke_label", "stroke_names"]
    for word in word_info:
        hw_word = word["handwriting"]
        gt_word = word["groundtruth"]
        print(json_utils.formatting(hw_word['content'] if 'content' in hw_word else ""))
        # print(json_utils.formatting(gt_info['content'] if 'content' in gt_info else ""))
        text = ["{}={}".format(key, hw_word[key]) for key in keys if key in hw_word]
        print("\t".join(text))
        if len(hw_word['stroke_segs']) == 0 or hw_word['mask'] is None: return
        concat_hw_gt_stroke_image(hw_mask=hw_word['mask'], hw_segs=hw_word['stroke_segs'],
                                  gt_mask=gt_word['mask'], gt_segs=gt_word['stroke_segs'], vis=True)
        for i in range(len(hw_word['piece_segs'])):
            hw_stroke = hw_word['stroke_segs'][i]
            hw_piece = hw_word['piece_segs'][i]
            gt_stroke = gt_word['stroke_segs'][i]
            gt_piece = gt_word['piece_segs'][i]
            names = hw_word['stroke_label'][i]
            print("i={:0=3d}\tname={}\tpiece_num={}".format(i, names, len(hw_piece)))
            if len(hw_piece) == 0: continue
            piece_image = concat_hw_gt_stroke_image(hw_mask=hw_stroke, hw_segs=hw_piece,
                                                    gt_mask=gt_stroke, gt_segs=gt_piece, vis=False)
            image_utils.cv_show_image("mask-seg-piece", piece_image)
        cv2.destroyWindow("mask-seg-piece")
        print("---" * 20)


def concat_packer(packers: List[List[Dict]]):
    if len(packers) == 0:
        return []
    if len(packers) == 1:
        return packers
    out_packers = packers[0]
    nums = len(packers[0])
    for p in range(1, len(packers)):
        assert nums == len(packers[p])
        for i in range(nums):
            out_packers[i].update(packers[p][i])
    return out_packers


def word_packer(unpacker: Dict, keys):
    packer = []
    for key in keys:
        if not packer:
            packer = [{key: v} for v in unpacker[key]]
        else:
            for i in range(len(unpacker[key])):
                packer[i][key] = unpacker[key][i]
    return packer


def word_unpacker(packer: List[Dict], keys):
    unpacker = {key: [] for key in keys}
    for i in range(len(packer)):
        for key, value in packer[i].items():
            if key in keys:
                unpacker[key].append(value)
    return unpacker


def get_content_value(content: Dict, key: str):
    """在字典content中,查找包含key的所有数据"""
    r = None
    for k, v in content.items():
        if k == key:
            return v
        elif isinstance(v, Dict):
            r = get_content_value(v, key)
        elif isinstance(v, List):
            r = [get_content_value(c, key) for c in v if isinstance(c, Dict)]
    return r


def find_key_metadata(metadata: List[Dict], key):
    """在metadata中,查找包含key的所有数据"""
    values = []
    for content in metadata:
        v = get_content_value(content, key)
        values.append(v)
    return values


def draw_word_packer_in_image(packer, image, vis=False, delay=0):
    for word in packer:
        image = draw_word_in_image(image, word["mask"], word["box"])
    if vis: image_utils.cv_show_image("image-mask", image, delay=delay)
    return image


def draw_word_in_image(image, mask, box):
    if not isinstance(image, np.ndarray): return image
    xmin, ymin, xmax, ymax = box
    mask: np.ndarray = image_utils.resize_image(mask, size=(xmax - xmin, ymax - ymin))
    if mask.ndim == 2: mask = cv2.cvtColor(mask, cv2.COLOR_GRAY2BGR)
    image = image_utils.cv_paste_image(image, mask, start_point=(int(xmin), int(ymin)))
    return image


def show_word_packer(packer, image, keys=[], split_line=False, color=(255, 0, 0), thickness=2, vis=False, delay=0):
    _keys = ['label', 'cls_score', 'box', 'det_score']
    mask = image.copy() if isinstance(image, np.ndarray) else None
    for word in packer:
        label = word['label'] if "label" in word else ""
        images = [word[k] for k in keys if k in word]
        info = ["{}:{}".format(k, word[k]) for k in _keys if k in word]
        print(info)
        if isinstance(image, np.ndarray):
            image = image_utils.draw_image_bboxes_text(image,
                                                       boxes=[word["box"]],
                                                       boxes_name=[label],
                                                       color=color,
                                                       thickness=thickness)
            mask = draw_word_in_image(mask, word["mask"], word["box"])
            if vis: show_images("dets", [image, mask], split_line=split_line, delay=1)
        if vis: show_images("packer", images, split_line=split_line, delay=delay)
    return image, mask


def show_word_unpacker(unpacker, image, keys=[], split_line=False, color=(255, 0, 0), thickness=2, vis=False, delay=0):
    _keys = ['label', 'cls_score', 'box', 'det_score']
    mask = image.copy() if isinstance(image, np.ndarray) else None
    for i in range(len(unpacker["box"])):
        label = unpacker["label"][i] if "label" in unpacker else ""
        images = [unpacker[k][i] for k in keys if k in unpacker]
        info = ["{}:{}".format(k, unpacker[k][i]) for k in _keys if k in unpacker]
        print(info)
        if isinstance(image, np.ndarray):
            image = image_utils.draw_image_bboxes_text(image,
                                                       boxes=[unpacker["box"][i]],
                                                       boxes_name=[label],
                                                       color=color,
                                                       thickness=thickness)
            mask = draw_word_in_image(mask, unpacker["mask"][i], unpacker["box"][i])
            if vis: show_images("dets", [image, mask], split_line=split_line, delay=1)
        if vis: show_images("unpacker", images, split_line=split_line, delay=delay)
    return image, mask


def show_images(title, images, use_rgb=False, split_line=False, delay=0):
    if isinstance(images, np.ndarray): images = [images]
    if len(images) == 0: return
    image = image_utils.image_hstack(images, split_line=split_line)
    image_utils.cv_show_image(title, image, use_rgb=use_rgb, delay=delay)


def save_images(images, dsize=None, out_dir="./output"):
    for img in images:
        name = file_utils.get_time("p") + ".png"
        file = file_utils.create_dir(out_dir, None, name)
        img = image_utils.resize_image(img, size=dsize)
        cv2.imwrite(file, img)


def save_packer_images(packer, keys=["crop"], dsize=None, out_dir="./output"):
    for i, word in enumerate(packer):
        time = file_utils.get_time("p")
        for k in keys:
            name = "{:0=3d}_{}_{}.png".format(i, time, k)
            file = file_utils.create_dir(out_dir, None, name)
            img = image_utils.resize_image(word[k], size=dsize)
            cv2.imwrite(file, img)



# -*-coding: utf-8 -*-
"""
    @Author : panjq
    @E-mail : pan_jinquan@163.com
    @Date   : 2022-06-29 18:58:33
    @Brief  :
"""
# -*- coding: utf-8 -*-
"""
# --------------------------------------------------------
# @Author : panjq
# @E-mail : pan_jinquan@163.com
# @Date   : 2020-02-05 11:01:49
# --------------------------------------------------------
"""
import os
import numpy as np
import cv2
import glob
import random


class Dataset(object):
    """
    from torch.utils.data import Dataset,DataLoader, ConcatDataset
    """

    def __init__(self, **kwargs):
        self.image_id = []

    def __getitem__(self, index):
        raise NotImplementedError

    def __add__(self, other):
        return ConcatDataset([self, other])

    def __len__(self):
        raise NotImplementedError

    @staticmethod
    def read_files(filename, split=None):
        """
        :param filename:
        :param split:分割
        :return:
        """
        image_id = []
        with open(filename, encoding="utf-8") as f:
            for line in f:
                line = line.rstrip().split(split) if split else line.rstrip()
                line = line[0] if isinstance(line, list) and len(line) == 1 else line
                image_id.append(line)
        return image_id

    @staticmethod
    def get_file_list(dir_root, postfix=['*.jpg'], basename=True):
        """
        获得文件列表
        :param dir_root: 图片文件目录
        :param postfix: 后缀名，可是多个如，['*.jpg','*.png']
        :param basename: 返回的列表是文件名（True），还是文件的完整路径(False)
        :return:
        """
        file_list = []
        for format in postfix:
            format = os.path.join(dir_root, format)
            image_list = glob.glob(format)
            if image_list:
                file_list += image_list
        file_list = sorted(file_list)
        if basename:
            file_list = [os.path.basename(f).split(".")[0] for f in file_list]
        return file_list

    @staticmethod
    def xyxy2cxcywh(xyxy, width=None, height=None, norm=False):
        """
        将(xmin, ymin, xmax, ymax)转换为(cx,cy,w,h)
        """
        cxcywh = np.zeros_like(xyxy)
        cxcywh[:, 0] = (xyxy[:, 2] + xyxy[:, 0]) / 2  # cx
        cxcywh[:, 1] = (xyxy[:, 3] + xyxy[:, 1]) / 2  # cy
        cxcywh[:, 2] = (xyxy[:, 2] - xyxy[:, 0])  # w
        cxcywh[:, 3] = (xyxy[:, 3] - xyxy[:, 1])  # h
        # xyxy = np.concatenate([cxcywh[..., :2] - cxcywh[..., 2:] / 2,
        #                        cxcywh[..., :2] + cxcywh[..., 2:] / 2], axis=1)
        if norm:
            cxcywh = cxcywh / (width, height, width, height)
        return cxcywh

    @staticmethod
    def cxcywh2xyxy(cxcywh, width=None, height=None, unnorm=False):
        """
        将(cx,cy,w,h)转换为(xmin, ymin, xmax, ymax)
        """
        xyxy = np.zeros_like(cxcywh)
        xyxy[:, 0] = cxcywh[:, 0] - cxcywh[:, 2] / 2  # top left x
        xyxy[:, 1] = cxcywh[:, 1] - cxcywh[:, 3] / 2  # top left y
        xyxy[:, 2] = cxcywh[:, 0] + cxcywh[:, 2] / 2  # bottom right x
        xyxy[:, 3] = cxcywh[:, 1] + cxcywh[:, 3] / 2  # bottom right y
        if unnorm:
            xyxy = xyxy * (width, height, width, height)
        return xyxy

    @staticmethod
    def clip_box(box, width, height):
        # xmin, ymin, xmax, ymax = bbox
        box[0] = max(0, box[0])
        box[1] = max(0, box[1])
        box[2] = min(width - 1, box[2])
        box[3] = min(height - 1, box[3])
        return box

    @staticmethod
    def check_box(box, width, height):
        xmin, ymin, xmax, ymax = box
        sw = (xmax - xmin) / width
        sh = (ymax - ymin) / height
        ok = True
        if sw < 0 or sw > 1:
            ok = False
        elif sh < 0 or sh > 1:
            ok = False
        return ok

    @staticmethod
    def search_path(root, sub_dir=[]):
        """搜索可能存在的子目录"""
        for sub in sub_dir:
            path = os.path.join(root, sub)
            if os.path.exists(path):
                return path
        return None


class ConcatDataset(Dataset):
    """ Concat Dataset """

    def __init__(self, datasets, shuffle=False):
        """
        import torch.utils.data as torch_utils
        voc1 = PolygonParser(filename1)
        voc2 = PolygonParser(filename2)
        voc=torch_utils.ConcatDataset([voc1, voc2])
        ====================================
        :param datasets:
        :param shuffle:
        """
        super(ConcatDataset, self).__init__()
        assert len(datasets) > 0, 'dataset should not be an empty iterable'
        # super(ConcatDataset, self).__init__()
        if not isinstance(datasets, list):
            datasets = [datasets]
        self.image_id = []
        self.dataset = datasets
        self.shuffle = shuffle
        for dataset_id, dataset in enumerate(self.dataset):
            image_id = dataset.image_id
            image_id = self.add_dataset_id(image_id, dataset_id)
            self.image_id += image_id
            self.classes = dataset.classes
        if shuffle:
            random.seed(200)
            random.shuffle(self.image_id)

    def add_dataset_id(self, image_id, dataset_id):
        """
        :param image_id:
        :param dataset_id:
        :return:
        """
        out_image_id = []
        for id in image_id:
            out_image_id.append({"dataset_id": dataset_id, "image_id": id})
        return out_image_id

    def __getitem__(self, index):
        """
        :param index: int
        :return:
        """
        dataset_id = self.image_id[index]["dataset_id"]
        image_id = self.image_id[index]["image_id"]
        dataset = self.dataset[dataset_id]
        data = dataset.__getitem__(image_id)
        return data

    def get_image_anno_file(self, index):
        dataset_id = self.image_id[index]["dataset_id"]
        image_id = self.image_id[index]["image_id"]
        return self.dataset[dataset_id].get_image_anno_file(image_id)

    def get_annotation(self, xml_file):
        return self.dataset[0].get_annotation(xml_file)

    def read_image(self, image_file):
        return self.dataset[0].read_image(image_file, use_rgb=self.dataset[0].use_rgb)

    def __len__(self):
        return len(self.image_id)



import cv2
import matplotlib.pyplot as plt
import numpy as np
import torch


def get_colormap(data_type="custom"):
    if data_type == 'pascal':
        num_classes = 21
        colormap = get_pascal_colormap()
    elif data_type == 'coco':
        num_classes = 21
        colormap = get_coco_colormap()
    elif data_type == 'custom':
        colormap = get_custom_colormap()
        num_classes = len(colormap)
    elif data_type == 'cityscapes':
        num_classes = 19
        colormap = get_cityscapes_colormap()
    else:
        raise NotImplementedError
    return colormap, num_classes


def decode_seg_map_sequence(label_masks, data_type='custom'):
    rgb_masks = []
    for label_mask in label_masks:
        rgb_mask = decode_segmap(label_mask, data_type)
        rgb_masks.append(rgb_mask)
    rgb_masks = torch.from_numpy(np.array(rgb_masks).transpose([0, 3, 1, 2]))
    return rgb_masks


def decode_segmap(label_mask, data_type="custom", plot=False):
    """Decode segmentation class labels into a color image
    Args:
        label_mask (np.ndarray): an (M,N) array of integer values denoting
          the class label at each spatial location.
        plot (bool, optional): whether to show the resulting color image
          in a figure.
    Returns:
        (np.ndarray, optional): the resulting decoded color image.
    """
    colormap, n_classes = get_colormap(data_type)
    r = label_mask.copy()
    g = label_mask.copy()
    b = label_mask.copy()
    for ll in range(0, n_classes):
        r[label_mask == ll] = colormap[ll, 0]
        g[label_mask == ll] = colormap[ll, 1]
        b[label_mask == ll] = colormap[ll, 2]
    rgb = np.zeros((label_mask.shape[0], label_mask.shape[1], 3), dtype=np.uint8)
    rgb[:, :, 0] = r
    rgb[:, :, 1] = g
    rgb[:, :, 2] = b
    if plot:
        plt.imshow(rgb)
        plt.show()
    return rgb


def decode_color_mask(mask, data_type="custom", plot=False) -> np.ndarray:
    """Decode segmentation class labels into a color image
    Args:
        mask (np.ndarray): an (M,N) array of integer values denoting
          the class label at each spatial location.
        plot (bool, optional): whether to show the resulting color image
          in a figure.
    Returns:
        (np.ndarray, optional): the resulting decoded color image.
    """
    colormap, n_classes = get_colormap(data_type)
    omask = mask.copy()
    if len(omask.shape) == 2:
        omask = cv2.cvtColor(omask, cv2.COLOR_GRAY2BGR)
    m = cv2.cvtColor(omask, cv2.COLOR_BGR2GRAY)
    for ll in range(0, n_classes):
        omask[m == ll] = colormap[ll]
    if plot:
        plt.imshow(omask)
        plt.show()
    return omask


def decode_color_image_mask(image, mask, data_type='custom'):
    """
    :param image: BGR原始图像
    :param mask: 对应的Mask
    :param data_type: 数据类型
    :return:
    """
    color_mask = decode_color_mask(mask, data_type=data_type, plot=False)
    alpha = mask.copy()
    alpha[mask > 0] = 255
    if len(alpha.shape) == 2:
        # alpha = cv2.cvtColor(alpha, cv2.COLOR_GRAY2BGR)
        alpha = alpha[:, :, np.newaxis]
    alpha = np.asarray(alpha / 255.0, dtype=np.float32)
    color_image = np.asarray(image / 2 + color_mask / 2, dtype=np.uint8)
    color_image = color_image * alpha + image * (1 - alpha)
    color_image = np.asarray(np.clip(color_image, 0, 255), dtype=np.uint8)
    return color_image, color_mask


def encode_segmap(mask):
    """Encode segmentation label images as pascal classes
    Args:
        mask (np.ndarray): raw segmentation label image of dimension
          (M, N, 3), in which the Pascal classes are encoded as colours.
    Returns:
        (np.ndarray): class map with dimensions (M,N), where the value at
        a given location is the integer denoting the class index.
    """
    mask = mask.astype(int)
    label_mask = np.zeros((mask.shape[0], mask.shape[1]), dtype=np.int16)
    for ii, label in enumerate(get_pascal_colormap()):
        label_mask[np.where(np.all(mask == label, axis=-1))[:2]] = ii
    label_mask = label_mask.astype(int)
    return label_mask


def get_cityscapes_colormap():
    return np.array([[128, 64, 128], [244, 35, 232], [70, 70, 70], [102, 102, 156], [190, 153, 153], [153, 153, 153],
                     [250, 170, 30], [220, 220, 0], [107, 142, 35], [152, 251, 152], [0, 130, 180], [220, 20, 60],
                     [255, 0, 0], [0, 0, 142], [0, 0, 70], [0, 60, 100], [0, 80, 100], [0, 0, 230], [119, 11, 32]])


def get_pascal_colormap():
    """Load the mapping that associates pascal classes with label colors
    Returns:
        np.ndarray with dimensions (21, 3)
    """
    return np.asarray([[0, 0, 0], [128, 0, 0], [0, 128, 0], [128, 128, 0],
                       [0, 0, 128], [128, 0, 128], [0, 128, 128], [128, 128, 128],
                       [64, 0, 0], [192, 0, 0], [64, 128, 0], [192, 128, 0],
                       [64, 0, 128], [192, 0, 128], [64, 128, 128], [192, 128, 128],
                       [0, 64, 0], [128, 64, 0], [0, 192, 0], [128, 192, 0],
                       [0, 64, 128]])


get_coco_colormap = get_pascal_colormap


def get_custom_colormap():
    """Load the mapping that associates pascal classes with label colors
    Returns:
        np.ndarray with dimensions (21, 3)
    """
    color_map = [(0, 0, 0), (255, 0, 0), (0, 255, 0), (0, 0, 255),
                 (128, 0, 0), (0, 128, 0), (128, 128, 0),
                 (0, 0, 128), (128, 0, 128), (0, 128, 128), (128, 128, 128),
                 (64, 0, 0), (192, 0, 0), (64, 128, 0), (192, 128, 0),
                 (64, 0, 128), (192, 0, 128), (64, 128, 128), (192, 128, 128),
                 (0, 64, 0), (128, 64, 0), (0, 192, 0), (128, 192, 0), (0, 64, 128)] * 10
    return np.asarray(color_map)



# -*-coding: utf-8 -*-
"""
    @Author : panjq
    @E-mail : pan_jinquan@163.com
    @Date   : 2021-09-08 16:58:52
"""

import math
import random
import torch
import cv2
import numpy as np


def clip_xyxy(xyxy: np.ndarray, valid_range):
    """
    限制xyxy的有效范围，避免越界
    :param xyxy:
    :param valid_range:有效范围(xmin,ymin,xmax,ymax)
    :return:
    """
    xmin, ymin, xmax, ymax = valid_range
    xyxy[:, [0, 2]] = np.clip(xyxy[:, [0, 2]], xmin, xmax)
    xyxy[:, [1, 3]] = np.clip(xyxy[:, [1, 3]], ymin, ymax)
    return xyxy


def xyxy2xywh(xyxy: np.ndarray):
    """(xmin,ymin,xmax,ymax)==>(xmin,ymin,w,h)"""
    if not isinstance(xyxy, np.ndarray): xyxy = np.asarray(xyxy)
    xywh = xyxy.copy()
    xywh[:, 2] = xywh[:, 2] - xywh[:, 0]  # w=xmax-xmin
    xywh[:, 3] = xywh[:, 3] - xywh[:, 1]  # w=ymax-ymin
    return xywh


def xywh2xyxy(xywh: np.ndarray):
    """(xmin,ymin,w,h)==>(xmin,ymin,xmax,ymax)"""
    if not isinstance(xywh, np.ndarray): xywh = np.asarray(xywh)
    xyxy = xywh.copy()
    xyxy[:, 2] = xyxy[:, 0] + xyxy[:, 2]  # xmax=xmin+w
    xyxy[:, 3] = xyxy[:, 1] + xyxy[:, 3]  # ymax=ymin+h
    return xyxy


def xyxy2cxcywh(xyxy: np.ndarray, width=None, height=None, normalized=False):
    """(xmin, ymin, xmax, ymax)==>(cx,cy,w,h)"""
    if not isinstance(xyxy, np.ndarray): xyxy = np.asarray(xyxy)
    cxcywh = xyxy.copy()
    cxcywh[:, 0] = (xyxy[:, 2] + xyxy[:, 0]) / 2  # cx
    cxcywh[:, 1] = (xyxy[:, 3] + xyxy[:, 1]) / 2  # cy
    cxcywh[:, 2] = (xyxy[:, 2] - xyxy[:, 0])  # w
    cxcywh[:, 3] = (xyxy[:, 3] - xyxy[:, 1])  # h
    if normalized:
        cxcywh[:, 0:4] = cxcywh[:, 0:4] / (width, height, width, height)
    return cxcywh


def cxcywh2xyxy(cxcywh: np.ndarray, width=None, height=None, normalized=False):
    """(cx,cy,w,h)==>xmin, ymin, xmax, ymax)"""
    if not isinstance(cxcywh, np.ndarray): cxcywh = np.asarray(cxcywh)
    xyxy = cxcywh.copy()
    xyxy[:, 0] = cxcywh[:, 0] - cxcywh[:, 2] / 2  # top left x
    xyxy[:, 1] = cxcywh[:, 1] - cxcywh[:, 3] / 2  # top left y
    xyxy[:, 2] = cxcywh[:, 0] + cxcywh[:, 2] / 2  # bottom right x
    xyxy[:, 3] = cxcywh[:, 1] + cxcywh[:, 3] / 2  # bottom right y
    if normalized:
        xyxy[:, 0:4] = xyxy[:, 0:4] * (width, height, width, height)
    return xyxy


def extend_xyxy(xyxy: np.ndarray, scale=[1.0, 1.0], valid_range=[]):
    """
    :param bboxes: [[xmin, ymin, xmax, ymax]]
    :param scale: [sx,sy]==>(W,H)
    :param valid_range:有效范围(xmin,ymin,xmax,ymax)
    :return:
    """
    if not isinstance(xyxy, np.ndarray): xyxy = np.asarray(xyxy)
    cxcywh = xyxy.copy()
    cxcywh[:, 0] = (xyxy[:, 2] + xyxy[:, 0]) / 2  # cx
    cxcywh[:, 1] = (xyxy[:, 3] + xyxy[:, 1]) / 2  # cy
    cxcywh[:, 2] = (xyxy[:, 2] - xyxy[:, 0]) * scale[0]  # w
    cxcywh[:, 3] = (xyxy[:, 3] - xyxy[:, 1]) * scale[1]  # h
    dxyxy = cxcywh2xyxy(cxcywh, width=None, height=None, normalized=False)
    if valid_range: dxyxy = clip_xyxy(dxyxy, valid_range=valid_range)
    return dxyxy


def extend_xywh(xywh: np.ndarray, scale=[1.0, 1.0]):
    """
    :param bboxes: [[xmin, ymin, xmax, ymax]]
    :param scale: [sx,sy]==>(W,H)
    :return:
    """
    if not isinstance(xywh, np.ndarray): xywh = np.asarray(xywh)
    xyxy = xywh2xyxy(xywh)
    xyxy = extend_xyxy(xyxy, scale)
    dxywh = xyxy2xywh(xyxy)
    return dxywh


def get_square_bboxes(boxes, use_max=True, baseline=-1):
    """
    将bboxes转换为正方形的bboxes
    :param boxes:
    :param use_max: 是否按照每个box(w,h)最大值进行转换
    :param baseline: 当baseline>0，表示正方形最小边长
    :return:
    """
    if not isinstance(boxes, np.ndarray): boxes = np.asarray(boxes)
    center = xyxy2cxcywh(boxes)
    if use_max:
        b = np.max(center[:, 2:4], axis=1)
    else:
        b = np.min(center[:, 2:4], axis=1)
    if baseline > 0:
        index = b < baseline
        b[index] = baseline
    b = b.reshape(len(boxes), 1)
    center[:, 2:4] = b
    _boxes = cxcywh2xyxy(center)
    return _boxes


def get_square_rects(rects, use_max=True, baseline=-1):
    """
    将rects转换为正方形的bboxes
    :param rects: xywh
    :param use_max: 是否按照每个box(w,h)最大值进行转换
    :param baseline: 当baseline>0，表示正方形最小边长
    :return:
    """
    boxes = xywh2xyxy(rects)
    boxes = get_square_bboxes(boxes, use_max=use_max, baseline=baseline)
    rects = xyxy2xywh(boxes)
    return rects


def get_section(start, end, nums=2, scale=1.0, dtype=None):
    """
    均匀划分nums个线段，并返回截断点(nums+1)
    s1 = get_section(start, end, nums=nums)
    等价于：
    s2 = np.linspace(start, end, num=nums + 1,endpoint=True)
    :param start: 起点
    :param end: 终点
    :param nums: 将范围[start,end]均匀划分的段数，默认2段
    :param scale: 对范围[start,end]进行缩放
    :param dtype: 输出类型
    :return: 返回截断点,其个数是nums+1
    """
    d = end - start
    c = (start + end) * 0.5  # 中心点
    start = int(c - d * scale * 0.5)
    end = int(c + d * scale * 0.5)
    unit = (end - start) / nums
    out = np.arange(start, end + 0.1 * unit, unit)
    dtype = dtype if dtype else out.dtype
    out = np.asarray(out, dtype=dtype)
    return out


def box_iou_v1(box1, box2):
    """
    :param box1: 预测框(n, 4)
    :param box2: GT框 (m, 4)
    :return: IOU (n, m)
    numpy 广播机制 从后向前对齐。 维度为1 的可以重复等价为任意维度
    eg: (4,3,2)   (3,2)  (3,2)会扩充为(4,3,2)
        (4,1,2)   (3,2) (4,1,2) 扩充为(4, 3, 2)  (3, 2)扩充为(4, 3,2) 扩充的方法为重复
    广播会在numpy的函数 如sum, maximun等函数中进行
    pytorch同理。
    扩充维度的方法：
    eg: a  a.shape: (3,2)  a[:, None, :] a.shape: (3, 1, 2) None 对应的维度相当于newaxis
    """
    if not isinstance(box1, np.ndarray):
        box1 = np.array(box1)
    if not isinstance(box2, np.ndarray):
        box2 = np.array(box2)
    lt = np.maximum(box1[:, None, :2], box2[:, :2])  # left_top (x, y)
    rb = np.minimum(box1[:, None, 2:], box2[:, 2:])  # right_bottom (x, y)
    wh = np.maximum(rb - lt + 1, 0)  # inter_area (w, h)
    inter_areas = wh[:, :, 0] * wh[:, :, 1]  # shape: (n, m)
    box_areas = (box1[:, 2] - box1[:, 0] + 1) * (box1[:, 3] - box1[:, 1] + 1)
    gt_areas = (box2[:, 2] - box2[:, 0] + 1) * (box2[:, 3] - box2[:, 1] + 1)
    iou = inter_areas / (box_areas[:, None] + gt_areas - inter_areas)
    return iou


def box_iou_v2(box1, box2):
    """
    https://www.jb51.net/article/178732.htm
    :param box1:
    :param box2:
    :return:
    """
    if not isinstance(box1, np.ndarray):
        box1 = np.array(box1)
    if not isinstance(box2, np.ndarray):
        box2 = np.array(box2)
    xmin1, ymin1, xmax1, ymax1, = np.split(box1, 4, axis=-1)
    xmin2, ymin2, xmax2, ymax2, = np.split(box2, 4, axis=-1)
    area1 = (xmax1 - xmin1) * (ymax1 - ymin1)
    area2 = (xmax2 - xmin2) * (ymax2 - ymin2)
    ymin = np.maximum(ymin1, np.squeeze(ymin2, axis=-1))
    xmin = np.maximum(xmin1, np.squeeze(xmin2, axis=-1))
    ymax = np.minimum(ymax1, np.squeeze(ymax2, axis=-1))
    xmax = np.minimum(xmax1, np.squeeze(xmax2, axis=-1))
    h = np.maximum(ymax - ymin, 0)
    w = np.maximum(xmax - xmin, 0)
    intersect = h * w
    union = area1 + np.squeeze(area2, axis=-1) - intersect
    iou = intersect / union
    return iou


class YOLOCoords(object):
    def __init__(self, max_boxes=120, norm=False):
        self.max_boxes = max_boxes
        self.norm = norm

    def __call__(self, image, boxes, labels):
        dboxes = np.zeros((self.max_boxes, 4))
        dlabels = np.zeros((self.max_boxes))
        if len(boxes) > 0:
            width, height, _ = image.shape
            cboxes = xyxy2cxcywh(boxes, width, height, normalized=self.norm)
            dboxes[0:len(cboxes)] = cboxes
            dlabels[0:len(cboxes)] = labels
        return image, dboxes, dlabels


def show_image(name, image, boxes, labels, center2bboxes=False, untranspose=False, waitKey=0):
    from pybaseutils import image_utils
    if center2bboxes:
        boxes = cxcywh2xyxy(boxes)
    if untranspose:
        image = image_utils.untranspose(image)
    image = image_utils.show_image_bboxes_text(name, image, boxes, labels, delay=waitKey)
    return image


def demo_for_augment():
    from pybaseutils import image_utils
    input_size = [320, 320]
    image_path = "test.jpg"
    boxes = [[98, 42, 160, 100], [244, 260, 297, 332], [98 + 50, 42 + 50, 160 + 50, 100 + 50]]
    labels = [1, 2, 3]
    image = image_utils.read_image(image_path)
    boxes = np.asarray(boxes, np.float32)
    labels = np.asarray(labels)
    augment = YOLOCoords(max_boxes=120, norm=False)
    for i in range(1000):
        boxes = extend_xyxy(boxes, scale=[1.2, 1.2])
        dst_image, dst_label, dst_boxes = augment(image, boxes.copy(), labels.copy())
        image = show_image("detd", image, dst_label, dst_boxes, center2bboxes=True, untranspose=False, waitKey=0)


if __name__ == "__main__":
    start, end = 1, 20
    nums = 5
    s1 = get_section(start, end, nums=5)
    s2 = np.linspace(start, end, num=nums + 1, endpoint=True)
    print(s1)
    print(s2)



# -*-coding: utf-8 -*-
"""
    @Author : panjq
    @E-mail : pan_jinquan@163.com
    @Date   : 2022-05-01 16:36:18
    @Url    : https://blog.csdn.net/xiaoyu_wu/article/details/102820384
"""

import threading
import time
from typing import List, Callable
from concurrent.futures import ThreadPoolExecutor, as_completed

# 创建线程锁
thread_lock = threading.Lock()


def thread_safety(func, *args, **kwargs):
    """
    使用以下加锁方式
    with 锁对象：
        do something

    和以下方式一样的效果
    lock.acquire()
    try:
        do something
    finnaly:
        lock.release()
    :param func:
    :param args:
    :param kwargs:
    :return:
    """
    with thread_lock:
        r = func(*args, **kwargs)
    return r


def consumer(image_path: str):
    """
    :param image_path:
    :return:
    """
    t = int(image_path.split(".")[0])
    time.sleep(t)
    # with thread_lock:
    print("正在处理数据：{}  ".format(image_path))
    return image_path


class ThreadPool(object):
    def __init__(self, max_workers=2):
        """
        :param max_workers:  线程池最大线程数
        :param maxsize:
        """
        self.executor = ThreadPoolExecutor(max_workers=max_workers)

    def submit(self, func: Callable, *args, **kwargs):
        """递交线程任务"""
        t = self.executor.submit(func, *args, **kwargs)
        return t

    def task_map(self, func: Callable, inputs: List):
        """线程任务，返回结果有序(map与submit的性能基本一致)"""
        # 通过executor的 map 获取已经完成的task的值
        result = []
        for r in self.executor.map(func, inputs):
            result.append(r)
        return result

    def task_submit(self, func: Callable, inputs: List):
        """线程任务，返回结果无序(map与submit的性能基本一致)"""
        task_list = [self.executor.submit(func, p) for p in inputs]
        result = []
        for task in as_completed(task_list):
            result.append(task.result())
        return result

    def task_submit_v1(self, func: Callable, inputs: List):
        """线程任务，返回结果无序"""
        task_list = [self.executor.submit(func, p) for p in inputs]
        result = []
        while len(task_list) > 0:
            index = []
            for i, task in enumerate(task_list):
                if task.done():
                    result.append(task.result())
                else:
                    index.append(i)
            task_list = [task_list[i] for i in index]
        return result

    def shutdown(self, wait=True):
        self.executor.shutdown(wait=wait)


def thread_lock_decorator():
    def decorator(func):
        def wrapper(*args, **kwargs):
            with thread_lock:
                result = func(*args, **kwargs)
            return result

        return wrapper

    return decorator


if __name__ == "__main__":
    import random
    from pybaseutils import debug

    tp = ThreadPool(max_workers=4)
    # contents = ["{}.jpg".format(i / 5) for i in range(10)]
    # random.shuffle(contents)
    contents = ["1.jpg", "4.jpg", "4.jpg", "4.jpg", "2.jpg"]
    print(contents)
    t0 = debug.TIME()
    result1 = tp.task_map(func=consumer, inputs=contents)
    t1 = debug.TIME()
    print()
    result2 = tp.task_submit(func=consumer, inputs=contents)
    t2 = debug.TIME()
    print("task_map   :{}".format(debug.RUN_TIME(t1 - t0)))
    print("task_submit:{}".format(debug.RUN_TIME(t2 - t1)))
    print(result1)
    print(result2)
